// A map between namespaces and base URLs for their online documentation
baseURLs = [
  ['GLib', 'https://docs.gtk.org/glib/'],
  ['Gio', 'https://docs.gtk.org/gio/'],
  ['GObject', 'https://docs.gtk.org/gobject/'],
  ['GModule', 'https://docs.gtk.org/gmodule/'],
]
